import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:logging/logging.dart';
import 'package:pana/pana.dart';
import 'package:path/path.dart' as p;
import 'package:pool/pool.dart';

final _clear = _isPosixTerminal ? '\x1b[2K\r' : '';
final _cyan = _isPosixTerminal ? '\u001b[36m' : '';
final _yellow = _isPosixTerminal ? '\u001b[33m' : '';
final _red = _isPosixTerminal ? '\u001b[31m' : '';
final _endColor = _isPosixTerminal ? '\u001b[0m' : '';
final _isPosixTerminal =
    !Platform.isWindows && stdioType(stdout) == StdioType.TERMINAL;

main() async {
  int cols;
  try {
    var result = Process.runSync('tput', ['cols']);
    cols = int.parse(result.stdout.trim());
  } catch (_) {
    // nevermind
  }

  var lastLine = true;
  int lastNumber;
  Logger.root.onRecord.listen((log) {
    var number = int.parse(log.loggerName, onError: (s) => null);

    if (number != null && log.error == null && log.level < Level.SEVERE) {
      if (number == lastNumber) {
        stderr.write(_clear);
      } else {
        stderr.writeln();
        lastNumber = number;
      }
      var message = [log.loggerName, log.level, log.message].join('\t');
      if (log.level >= Level.WARNING) {
        message = "$_red$message$_endColor";
      }

      if (cols != null && message.length > cols) {
        message = message.substring(0, cols);
      }

      stderr.write(message);
      lastLine = false;
      lastNumber = number;
    } else {
      if (!lastLine) {
        stderr.writeln();
        lastLine = true;
      }
      stderr.writeln([log.loggerName, log.level, log.message].join('\t'));
      if (log.error != null) {
        stderr.writeln(log.error);
      }
      if (log.stackTrace != null) {
        stderr.writeln(log.stackTrace);
      }
    }
  });

  // dockerName
  var dockerName = 'results';

  // docker build
  await _runProc('docker', [
    'build',
    '-t',
    dockerName,
    Directory.current.resolveSymbolicLinksSync()
  ]);

  var packages = new File('input_packages.txt').readAsLinesSync();

  var pool = new Pool(Platform.numberOfProcessors ~/ 2);

  print("Hacking through ${packages.length} package(s).");
  var count = 0;
  await Future.wait(packages.map((pkg) async {
    var resource = await pool.request();

    var logger = new Logger((++count).toString());
    try {
      // do things here!
      var result = await _runProc('docker', ['run', '--rm', dockerName, pkg],
          logger: logger);
      result = prettyJson(JSON.decode(result));
      await _writeResult(dockerName, pkg, result, false);
    } catch (e, stack) {
      logger.severe("Oops!", e, stack);
      await _writeResult(dockerName, pkg, [e, stack].join('\n'), true);
    } finally {
      resource.release();
    }
  }));
}

final _tempDirFind = new RegExp(r'/pana\.\d+\.[a-zA-Z0-9]+');

_writeResult(String dockerName, String pkg, String output, bool isError) async {
  var dir = new Directory(dockerName);
  if (!dir.existsSync()) {
    dir.createSync(recursive: true);
  }

  // cleanup random pana path fun
  output = output.replaceAll(_tempDirFind, '/pana');

  var name = isError ? '$pkg.error.txt' : '$pkg.json';
  var file = new File(p.join(dockerName, name));

  await file.writeAsString(output, mode: WRITE_ONLY, flush: true);
}

Future<String> _runProc(String proc, List<String> args, {Logger logger}) async {
  logger ??= Logger.root;

  var stdoutLines = <String>[];
  var stderrLines = <String>[];

  var process = await Process.start(proc, args);

  var items = await Future.wait(<Future<Object>>[
    process.exitCode,
    byteStreamSplit(process.stdout).forEach((outLine) {
      stdoutLines.add(outLine);
      logger.info(outLine);
    }),
    byteStreamSplit(process.stderr).forEach((errLine) {
      stderrLines.add(errLine);
      logger.warning(errLine);
    })
  ]);

  var exitCode = items.first as int;
  if (exitCode != 0) {
    throw stderrLines.join('\n');
  }

  return stdoutLines.join('\n');
}
